import zlib
# import pkw
# import io
# import gzip
# import bz2
import time
# import os


class NotMpqArchiveException(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return str(self.value)


class MpqFile:
    def __init__(self):
        self.fileName = ""
        self.fullFileName = ""
        self.blockIndex = 0
        self.hashIndex = 0
        self.offsetTable = []
        self.sectors = []
        self.dataBytes = []


class BlockTable:
    def __init__(self):
        self.filePos = 0
        self.c_size = 0
        self.f_size = 0
        self.flags = 0


class HashTable:
    def __init__(self):
        self.name1 = 0
        self.name2 = 0
        self.blockIndex = 0
        self.names = {}


class MpqHeader:
    def __init__(self):
        self.HeaderSize = 0
        self.ArchiveSize = 0
        self.FormatVersion = 0
        self.SectorSizeShift = 0
        self.HTOffset = 0
        self.BTOffset = 0
        self.HTEntries = 0
        self.BTEntries = 0


class Mpq:
    def __init__(self, archive_file, listfile=0):
        # cryptography constants
        self.MPQ_KEY_HASH_TABLE = 0xC3AF3770
        self.MPQ_KEY_BLOCK_TABLE = 0xEC83B3A3
        self.MPQ_HASH_TABLE_INDEX = 0x00000000
        self.MPQ_HASH_NAME_A = 0x00000100
        self.MPQ_HASH_NAME_B = 0x00000200
        self.MPQ_HASH_FILE_KEY = 0x00000300
        self.MPQ_HASH_KEY2_MIX = 0x00000400
        self.STORM_BUFFER_SIZE = 0x00000500

        # file flags
        self.MPQ_FILE_IMPLODE = 0x00000100  # Implode method (By PKWARE Data Compression Library)
        self.MPQ_FILE_COMPRESS = 0x00000200  # Compress methods (By multiple methods)
        self.MPQ_FILE_ENCRYPTED = 0x00010000  # Indicates whether file is encrypted
        self.MPQ_FILE_FIX_KEY = 0x00020000  # File decryption key has to be fixed
        # self.MPQ_FILE_PATCH_FILE = 0x00100000 The file is a patch file. Raw file data begin with TPatchInfo structure
        # self.MPQ_FILE_DELETE_MARKER = 0x02000000  #File is a deletion marker.
        self.MPQ_FILE_EXISTS = 0x80000000  # Set if file exists, reset when the file was deleted

        # compresion flags
        self.MPQ_COMPRESSION_ADPCM_MONO = 0x00000040
        self.MPQ_COMPRESSION_HUFFMANN = 0x00000001
        self.MPQ_COMPRESSION_ZLIB = 0x00000002
        self.MPQ_COMPRESSION_PKWARE = 0x00000008
        self.MPQ_COMPRESSION_BZIP2 = 0x00000010
        # self.MPQ_COMPRESSION_SPARSE = 0x00000020  #Sparse compression (added in Starcraft 2)
        # self.MPQ_COMPRESSION_LZMA = 0x00000012  #LZMA compression. Added in Starcraft 2.
        # self.MPQ_COMPRESSION_NEXT_SAME = 0xFFFFFFFF  #Same compression

        self.size = 0
        self.source = archive_file
        self.raw_bytes = b""
        self.stormBuffer = [0] * self.STORM_BUFFER_SIZE
        self.header = MpqHeader()
        self.HOffset = 0
        self.raw_ht = []
        self.raw_bt = []
        self.hash_table = {}
        self.block_table = []
        self.files = {}
        self.listfile = []
        self.temp_offset_table = []

        t = time.time()
        self.mpq_cryptography_init()
        # print("cryptography", time.time() - t)
        self.header_init()
        # print("header", time.time() - t)
        self.hash_table_init()
        # print("hash_table", time.time() - t)
        self.block_table_init()
        # print("block_table", time.time() - t)
        # if listfile:
        #    self.listfile_init(listfile)
        #    self.mpq_files_init()

    def encrypt_file(self, file, key):
        ot_len = int.from_bytes(file[:4], "little")
        ot = [int.from_bytes(file[i * 4: i * 4 + 4], "little") for i in range(ot_len)]
        for i in range(ot_len):
            self.encrypt_mpq_block(file[ot[i - 1]: ot[i]], i, key)

    def compress_file(self, file, encrypt=False, key=0):
        sectors = b""
        offsets_table = [0]
        sector_size = 512 * 2 ** self.header.SectorSizeShift
        # print(sector_size)
        f = file[:]

        while len(f) > 0:
            s = f[:sector_size]
            # print(len(s))
            sector = bytes([self.MPQ_COMPRESSION_ZLIB]) + zlib.compress(s, 9)
            # print(len(sector))
            if len(sector) >= len(s):
                sector = s[:]
                # print(sector)

            if encrypt:
                last = sector[len(sector) // 4 * 4:]
                sector = [int.from_bytes(sector[j * 4: j * 4 + 4], "little") for j in range(len(sector) // 4)]
                self.encrypt_mpq_block(sector, len(sector), key + len(offsets_table))
                sector = self.split_to_bytes(sector, last)
            offsets_table.append(offsets_table[-1] + len(sector))
            sectors += sector
            f = f[sector_size:]

        for i in range(len(offsets_table)):
            offsets_table[i] += (len(offsets_table)) * 4
        # print(offsets_table)
        if encrypt:
            self.encrypt_mpq_block(offsets_table, len(offsets_table), key)
        return self.split_to_bytes(offsets_table) + sectors

    def create_tables(self, ht, bt, db, file, name):
        if isinstance(file, dict):
            for i in file:
                self.create_tables(ht, bt, db, file[i], name + "\\" + i)
        else:
            name = name[1:]
            i = self.hash_string(name, self.MPQ_HASH_TABLE_INDEX) % len(ht)
            while ht[i] != 0:
                i = (i + 1) % len(ht)
            ht[i] = [self.hash_string(name, self.MPQ_HASH_NAME_A),
                     self.hash_string(name, self.MPQ_HASH_NAME_B),
                     0, len(bt) - 1]

            original_size = len(file)
            file = self.compress_file(file)
            bt.append([bt[-1][0] + bt[-1][1], len(file), original_size, self.MPQ_FILE_EXISTS | self.MPQ_FILE_COMPRESS])
            db += file

    def key_breaker(self, data, size):
        n = ((len(data) * 4 ^ data[0]) - 0xEEEEEEEE) & 0xFFFFFFFF
        # print(n)
        for i in range(0x100):
            key = (n - self.stormBuffer[self.MPQ_HASH_KEY2_MIX + i]) & 0xFFFFFFFF
            if key & 0xFF == i:
                #print(key)
                j = data[:]
                self.decrypt_mpq_block(j, len(j), key)
                #print(j)
                if j[-1] == size:
                    #print(key)
                    return key

    def is_file_exist(self, file_name):
        try:
            a = self.hash_string(file_name, self.MPQ_HASH_NAME_A)
            b = self.hash_string(file_name, self.MPQ_HASH_NAME_B)
            return self.hash_table[a][b]
        except KeyError:
            return -1

    def find_file(self, file_name="", block_index=-1):
        sectors = []
        data_bytes = []
        if block_index == -1:
            block_index = self.is_file_exist(file_name)
            if block_index == -1:
                return False

        # print(block_index,
        #      hex(self.block_table[block_index].filePos),
        #      hex(self.block_table[block_index].c_size),
        #      hex(self.block_table[block_index].f_size),
        #      hex(self.block_table[block_index].flags))
        # print("csize :", self.block_table[block_index].cSize)

        # print(file_name)
        # print(hex(block_index.flags))

        flags = self.block_table[block_index].flags
        c_size = self.block_table[block_index].c_size
        f_size = self.block_table[block_index].f_size
        # print(hex(flags))

        if self.block_table[block_index].flags & self.MPQ_FILE_COMPRESS:
            file_size = self.block_table[block_index].c_size
        else:
            file_size = self.block_table[block_index].f_size

        current_pos = self.HOffset + self.block_table[block_index].filePos
        self.source.seek(current_pos)
        # print("file offset", hex(current_pos))

        if self.block_table[block_index].flags == self.MPQ_FILE_EXISTS:
            data_bytes = self.source.read(self.block_table[block_index].f_size)
            if len(data_bytes) == 0:
                print("MPQ: Пустой файл")
                return False
        else:
            if self.block_table[block_index].flags & self.MPQ_FILE_ENCRYPTED:
                raw_bytes = self.source.read(file_size)  # self.blockTable[block_index].cSize)
                # print(len(raw_bytes))
                if len(raw_bytes) == 0:
                    print("MPQ: Пустой файл")
                    return False
                # print(0)
                offset_table_len = round(f_size / (512 * 2 ** self.header.SectorSizeShift) + 0.5) + 1
                # print(offset_table_len)
                offsets_table = [int.from_bytes(raw_bytes[i * 4: i * 4 + 4], "little") for i in range(offset_table_len)]
                # print("ot", offsets_table)
                if file_name != "":
                    key = self.hash_string(self.get_file_name(file_name), self.MPQ_HASH_FILE_KEY)
                    if self.block_table[block_index].flags & self.MPQ_FILE_FIX_KEY:
                        key = (key + self.block_table[block_index].filePos) ^ self.block_table[block_index].f_size
                else:
                    # print(offsets_table, hex(self.blockTable[block_index].cSize))
                    key = self.key_breaker(offsets_table, self.block_table[block_index].c_size)
                    # print(key)
                    if key is None:
                        print("MPQ: Не удалось подобрать ключ", block_index)
                        return False
                    else:
                        key += 1
                key &= 0xFFFFFFFF
                # print("key :", key)
                self.decrypt_mpq_block(offsets_table, offset_table_len, key - 1)
                # print(offsets_table)
                for i in range(1, len(offsets_table)):
                    sectors.append(raw_bytes[offsets_table[i - 1]: offsets_table[i]])
                for i in range(len(sectors)):
                    l0 = len(sectors[i])
                    l1 = l0 >> 2
                    sector = [int.from_bytes(sectors[i][j * 4: j * 4 + 4], "little") for j in range(l1)]
                    last_bytes = sectors[i][l1 * 4: l0]
                    self.decrypt_mpq_block(sector, l1, key + i)
                    sectors[i] = self.split_to_bytes(sector, last_bytes)
            else:
                # print(self.source.read(4))
                offsets_table = [self.read_c_int(1, current_pos)]
                current_pos += 4

                if offsets_table[0] > self.size:
                    print("MPQ: 123")
                    print(block_index, hex(flags))
                    return False

                for _ in range(offsets_table[0] // 4 - 1):
                    offsets_table.append(self.read_c_int(1, current_pos))
                    current_pos += 4

                for i in range(1, len(offsets_table)):
                    bytes_len = offsets_table[i] - offsets_table[i - 1]
                    sectors.append(self.read_c_int(2, current_pos, bytes_len))
                    # print(sectors[i - 1])
                    current_pos += bytes_len

            # print(sectors)
            if self.block_table[block_index].flags & self.MPQ_FILE_COMPRESS:
                # print(hex(sum(len(i) for i in sectors) + offset_table_len * 4))
                # fl = open("test_file", "wb")
                # print(offsets_table)
                for i in range(len(sectors)):
                    sector = sectors[i]
                    # print(sector)
                    # print(offsets_table[i + 1] - offsets_table[i], len(sector))
                    # if (offsets_table[i + 1] - offsets_table[i]) == len(sector):

                    if f_size == len(sector):
                        data_bytes += sector
                    elif sector[0] == self.MPQ_COMPRESSION_ZLIB:
                        data_bytes += zlib.decompress(sector[1:])
                        # print(1)
                    elif sector[0] | self.MPQ_COMPRESSION_ADPCM_MONO and sector[0] | self.MPQ_COMPRESSION_HUFFMANN:
                        # todo
                        # print(2)
                        pass
                    elif sector[0] == self.MPQ_COMPRESSION_PKWARE:
                        # print(i, block_index)
                        # print(sector[1:])
                        # data_bytes += pkw.decompress(sector[1:])
                        print("MPQ: Неизвестный формат сжатия")
                    else:
                        print("MPQ: Неизвестный формат сжатия")
                        # data_bytes += sector
                # fl.close()
            else:
                for sector in sectors:
                    data_bytes += sector

        return bytes(data_bytes)

    def file_type_check(self, block_index):
        self.source.seek(self.HOffset + self.block_table[block_index].filePos)
        if self.block_table[block_index].flags & self.MPQ_FILE_ENCRYPTED:
            offse_table_len = round(self.block_table[block_index].fSize / self.header.SectorSizeShift + 0.5) + 1
            raw_bytes = self.source.read(offse_table_len)
            offsets_table = [int.from_bytes(raw_bytes[i * 4: i * 4 + 4], "little") for i in range(offse_table_len)]
            pass

    @staticmethod
    def get_file_name(full_name):
        if "\\" in full_name:
            return full_name.split("\\")[-1]
        elif "/" in full_name:
            return full_name.split("/")[-1]
        else:
            return full_name

    def read_c_int(self, mode=0, pos=0, bytes_len=4):
        if mode == 0:
            return int.from_bytes(self.source.read(bytes_len), 'little')
        elif mode == 1:
            return int.from_bytes(self.raw_bytes[pos: pos + bytes_len], 'little')
        else:
            return self.raw_bytes[pos: pos + bytes_len]

    @staticmethod
    def split_to_bytes(a, last_bytes=b"", lenght=4):
        b = []
        for i in a:
            for _ in range(lenght):
                b.append(i & 0xFF)
                i >>= 8
        b += last_bytes
        # print(b, last_bytes)
        return bytes(b)

    def mpq_files_init(self):
        self.files = {}
        for f in self.listfile:
            block_index = self.is_file_exist(f)
            if block_index >= 0:
                try:
                    self.files[f] = self.find_file(f.lower().split("\\")[-1], block_index)
                except KeyError:
                    pass
                """folders = f.lower().split("\\")
                r = self.files
                for i in range(len(folders) - 1):
                    if not folders[i] in r:
                        r[folders[i]] = {}
                    r = r[folders[i]]
                try:
                    r[folders[-1]] = self.find_file(folders[-1], block_index)
                    #print(f)
                except KeyError:
                    pass"""

    def listfile_init(self, listfile):
        self.listfile = listfile.read().split("\n")

    def block_table_init(self):
        # print(self.header.BlockTableEntries)
        self.source.seek(self.HOffset + self.header.BTOffset)
        b = self.source.read(self.header.BTEntries * 16)
        b = [int.from_bytes(b[i * 4: i * 4 + 4], "little") for i in range(self.header.BTEntries * 4)]
        self.decrypt_mpq_block(b, self.header.BTEntries * 4, self.MPQ_KEY_BLOCK_TABLE)
        self.raw_bt = b[:]
        for i in range(self.header.BTEntries):
            bl = BlockTable()
            bl.filePos = self.to_c_int(b[i * 4])
            bl.c_size = b[i * 4 + 1]
            bl.f_size = b[i * 4 + 2]
            bl.flags = b[i * 4 + 3]

            if bl.filePos > self.size or -bl.filePos > self.HOffset:
                break

            if not bl.flags & self.MPQ_FILE_EXISTS:
                break

            """if bl.flags & self.MPQ_FILE_COMPRESS:
                if bl.cSize == 0:
                    continue
                if 0 < self.headerOffset + bl.filePos + bl.cSize < self.size:
                    self.blockTable[i] = bl
            else:
                if bl.fSize == 0:
                    continue
                if 0 < self.headerOffset + bl.filePos + bl.fSize < self.size:
                    self.blockTable[i] = bl"""
            #   if bl.flags & self.MPQ_FILE_EXISTS:
            self.block_table.append(bl)

            #    print(i, hex(bl.filePos), hex(bl.cSize), hex(bl.fSize), hex(bl.flags))
            #    print("3", time.time() - t)

        # print(len(self.blockTable))

    def hash_table_init(self):
        ht_start_pos = self.HOffset + self.header.HTOffset
        # print("hto", hex(ht_start_pos))
        self.header.HTEntries = min(self.header.HTEntries, (self.size - ht_start_pos) // 16)
        if self.header.BTOffset < 0:
            self.header.BTEntries = min(self.header.BTEntries, -self.header.BTOffset // 16)
        else:
            bt_start_pos = self.HOffset + self.header.BTOffset
            self.header.BTEntries = min(self.header.BTEntries, (self.size - bt_start_pos) // 16)
        # print("bte", self.HOffset, self.header.BTOffset, self.header.BTEntries)

        self.source.seek(ht_start_pos)
        b = self.source.read(self.header.HTEntries * 16)
        b = [int.from_bytes(b[i * 4: i * 4 + 4], "little") for i in range(self.header.HTEntries * 4)]
        # print(len(b))
        # print(len(b))
        self.decrypt_mpq_block(b, self.header.HTEntries * 4, self.MPQ_KEY_HASH_TABLE)
        self.raw_ht = b[:]

        # max_ht_index = 0
        max_bt_index = 0
        ht = []
        # bt_indexes = []
        for i in range(self.header.HTEntries):
            if b[i * 4] == 0xFFFFFFFF or b[i * 4] == 0xFFFFFFFE:
                continue

            if b[i * 4 + 3] > self.header.BTEntries:
                continue

            if b[i * 4 + 2] != 0:
                continue

            """
            #if b[i * 4] == self.hash_string("war3map.j", self.MPQ_HASH_NAME_A):
            #    if b[i * 4 + 1] == self.hash_string("war3map.j", self.MPQ_HASH_NAME_B):
            #        print("blockid", b[i * 4 + 3])

            #if b[i * 4 + 3] == 111:
                #print(0)

            #print(b[i * 4 + 3])
            # bt_indexes.append(b[i * 4 + 3])
            """

            ht.append(b[i * 4: i * 4 + 4])
            if b[i * 4 + 3] > max_bt_index:
                max_bt_index = b[i * 4 + 3]

            if not b[i * 4] in self.hash_table:
                self.hash_table[b[i * 4]] = {}
                self.hash_table[b[i * 4]][b[i * 4 + 1]] = b[i * 4 + 3]
            max_ht_index = i

        # print(n)
        # print("max i:", maxi)
        # self.bt_indexes = set(bt_indexes)
        # print(len(bt_indexes), max(bt_indexes) + 1, bt_indexes)
        # self.header.HashTableEntries = min(self.header.HashTableEntries, max_ht_index + 1)
        # print(self.header.HashTableEntries)
        # self.header.BlockTableEntries = min(self.header.BlockTableEntries, max(bt_indexes) + 1)
        # print(self.header.BlockTableEntries)
        return ht

    @staticmethod
    def to_c_int(a):
        b = a - (1 << 32) * (a >> 31)
        # print(b, hex(a), a)
        return b

    def header_init(self):
        self.raw_bytes = self.source.read()
        self.size = len(self.raw_bytes)
        # self.source.close()
        # print(self.size)
        # self.source = io.BytesIO(self.raw_bytes)
        self.source.seek(0)

        while True:
            a = self.source.read(4)
            if a == b'MPQ\x1A':
                break
            if not a:
                raise NotMpqArchiveException("Неверный формат файла")
            self.HOffset += 0x200
            self.source.read(508)

        # print(self.HOffset)
        # print(hex(self.headerOffset))
        self.header.HeaderSize = int.from_bytes(self.source.read(4), 'little')
        self.header.ArchiveSize = int.from_bytes(self.source.read(4), 'little')
        self.header.FormatVersion = int.from_bytes(self.source.read(2), 'little')
        # self.header.SectorSizeShift = 512 * 2 ** int.from_bytes(self.source.read(2), 'little')
        self.header.SectorSizeShift = int.from_bytes(self.source.read(2), 'little')
        self.header.HTOffset = Mpq.to_c_int(int.from_bytes(self.source.read(4), 'little'))
        self.header.BTOffset = Mpq.to_c_int(int.from_bytes(self.source.read(4), 'little'))
        self.header.HTEntries = int.from_bytes(self.source.read(4), 'little')
        self.header.BTEntries = int.from_bytes(self.source.read(4), 'little')
        # print(self.headerOffset)
        # print(self.header.HashTableOffset)
        # print(self.header.BlockTableOffset)
        # print(self.header.HashTableEntries)
        # print(self.header.BlockTableEntries)

    def mpq_cryptography_init(self):
        seed = 0x00100001
        for index1 in range(0x100):
            index2 = index1
            for _ in range(5):
                seed = (seed * 125 + 3) % 0x2AAAAB
                temp1 = (seed & 0xFFFF) << 0x10
                seed = (seed * 125 + 3) % 0x2AAAAB
                temp2 = (seed & 0xFFFF)
                self.stormBuffer[index2] = (temp1 | temp2)
                index2 += 0x100

    @staticmethod
    def to_bytes(n):
        b = bytes()
        for i in range(4):
            bytes += n % 0x100
            n //= 0x100
        return b

    def encrypt_mpq_block(self, data_block, length, key1):
        key2 = 0xEEEEEEEE
        for i in range(length):
            key2 = (key2 + self.stormBuffer[self.MPQ_HASH_KEY2_MIX + (key1 & 0xFF)]) & 0xFFFFFFFF
            dw_value = data_block[i]
            data_block[i] = (data_block[i] ^ (key1 + key2)) & 0xFFFFFFFF
            key1 = (((~key1 << 0x15) + 0x11111111) | (key1 >> 0xB)) & 0xFFFFFFFF
            key2 += dw_value + (key2 << 5) + 3

    def decrypt_mpq_block(self, data_block, length, key1):
        key2 = 0xEEEEEEEE
        for i in range(length):
            key2 = (key2 + self.stormBuffer[self.MPQ_HASH_KEY2_MIX + (key1 & 0xFF)]) & 0xFFFFFFFF
            data_block[i] = (data_block[i] ^ (key1 + key2)) & 0xFFFFFFFF
            key1 = (((~key1 << 0x15) + 0x11111111) | (key1 >> 0xB)) & 0xFFFFFFFF
            key2 = (key2 + data_block[i] + (key2 << 5) + 3) & 0xFFFFFFFF

    def hash_string(self, file_name, hash_type):
        seed1 = 0x7FED7FED
        seed2 = 0xEEEEEEEE

        s = [i for i in file_name.encode("utf-8")]
        for i in range(len(s)):
            if 97 <= s[i] <= 122:
                s[i] -= 32
            if s[i] == 47:
                s[i] = 92
        #print(s)
        for ch in s:
            #ch = ord(ch)
            seed1 = self.stormBuffer[hash_type + ch] ^ ((seed1 + seed2) & 0xFFFFFFFF)
            seed2 = (ch + seed1 + seed2 + (seed2 << 5) + 3) & 0xFFFFFFFF
            #print(hex(seed1 & 0xFF), hex(seed2 & 0xFF))
        return seed1
